{***************************************************************************}
{                            UNIDAD DE PINTURA 2

                                  Copyright (c) 1998 Marco Antonio Alvarado

Utiliza el modo $0101 640 * 480, 256 colores de VESA SVBE 1.2.

ADVERTENCIAS

* Evite usar esta unidad junto con la Unidad de Pintura 1.
* Algunas rutinas provocan efectos indeseables, como grficos despedazados,
  si se las llama desde interrupciones. Esto se debe a que alteran la
  posicin de la ventana cuando se est utilizando.}
{***************************************************************************}

{$x+,t-,g+}
unit upintur2;

interface

  uses
    umemoria,  {Unidad para Memoria}
    uvga,  {Unidad para Video Graphic Array}
    usvbe,  {Unidad para VESA Super VGA BIOS Extension}
    umd;  {Unidad para Mouse Driver}

  type

    rfiltro = record
      coeficientes : array [0..24] of integer;
      factor : integer;
      incremento : integer;
    end;

    rpunto2 = record
      x,
      y : integer;
    end;

    rpunto3 = record
      x,
      y,
      z : integer;
    end;

    rcuadro = record
      pintura : pointer;
      ancho,
      alto : word;
    end;

  const

    filtroborrosidad : rfiltro =
      (coeficientes :
        (1, 1, 1, 1, 1,
         1, 1, 1, 1, 1,
         1, 1, 1, 1, 1,
         1, 1, 1, 1, 1,
         1, 1, 1, 1, 1);
       factor : 25;
       incremento : 0);

    filtrosuavidad : rfiltro =
      (coeficientes :
        (0, 0, 0, 0, 0,
         0, 1, 3, 1, 0,
         0, 3, 5, 3, 0,
         0, 1, 3, 1, 0,
         0, 0, 0, 0, 0);
       factor : 21;
       incremento : 0);

    filtromareo : rfiltro =
      (coeficientes :
        (1, 0, 0, 0, 0,
         1, 0, 0, 0, 0,
         0, 0, 1, 0, 0,
         0, 0, 0, 0, 1,
         0, 0, 0, 0, 1);
       factor : 5;
       incremento : 0);

    filtroexpansion : rfiltro =
      (coeficientes :
        (1, 0, 0, 0, 1,
         0, 0, 0, 0, 0,
         0, 0, 0, 0, 0,
         0, 0, 0, 0, 0,
         1, 0, 0, 0, 1);
       factor : 4;
       incremento : 0);

    filtroproyeccion : rfiltro =
      (coeficientes :
        (1, 0, 0, 0, 0,
         0, 1, 0, 0, 0,
         0, 0, 1, 0, 0,
         0, 0, 1, 0, 0,
         0, 0, 1, 0, 0);
       factor : 5;
       incremento : 0);

    filtrofuego : rfiltro =
      (coeficientes :
        (0, 0, 0, 0, 0,
         0, 0, 0, 0, 0,
         0, 0, 1, 0, 0,
         0, 1, 1, 1, 0,
         0, 0, 0, 0, 0);
       factor : 4;
       incremento : 0);

  var
{Error:
  1 = el procesador no es 80386 o superior.
  2 = no puede poner el modo SVGA.}
    error : word;

    fundicion : pointer;

    filtro : rfiltro;

    masx1simbolo,
    masy1simbolo,
    masx2simbolo,
    masy2simbolo,
    masx3simbolo,
    masy3simbolo : integer;

    xraton,
    yraton : integer;
    botonizquierdoraton,
    botonderechoraton : boolean;

{/// Base //////////////////////////////////////////////////////////////////}

  function copiapuntobase (punto : longint) : byte;
  procedure pegapuntobase (color : byte; punto : longint);
  procedure llenapuntobase (color : byte; punto, cantidad : longint);
  procedure copiapinturabase (punto : longint; hacia : pointer; cantidad :
    longint);
  procedure pegapinturabase (desde : pointer; punto, cantidad : longint);
  procedure pegapinturahuecobase (desde : pointer; punto, cantidad : longint;
    color : byte);
  procedure pegapinturasolobase (desde : pointer; punto, cantidad : longint;
    color : byte);

{/// Pgina ////////////////////////////////////////////////////////////////}

  function tomapaginas : word;
  procedure ponepaginaactiva (pagina : word);
  function tomapaginaactiva : word;
  procedure ponepaginavisible (pagina : word);
  function tomapaginavisible : word;
  procedure regresapaginaactiva;
  procedure avanzapaginaactiva;
  procedure regresapaginavisible;
  procedure avanzapaginavisible;

{/// Plantilla /////////////////////////////////////////////////////////////}

  procedure ajustaplantilla;
  procedure tomaplantilla (var x1, y1, x2, y2 : integer);
  function anchoplantilla : word;
  function altoplantilla : word;
  procedure poneplantilla (x1, y1, x2, y2 : integer);
  procedure mueveplantilla (masx, masy : integer);
  procedure creceplantilla (masx, masy : integer);
  function puntodentroplantilla (x, y : integer) : boolean;
  function cajadentroplantilla (x1, y1, x2, y2 : integer) : boolean;
  function cuadrodentroplantilla (x, y : integer; ancho, alto : word) :
    boolean;

{/// Pintura ///////////////////////////////////////////////////////////////}

  function copiapunto (x, y : integer) : byte;
  procedure pegapunto (color : byte; x, y : integer);
  procedure pegalinea (color : byte; x1, y1, x2, y2 : integer);
  procedure pegalineas (color : byte; puntos : array of rpunto2; lineas :
    word);
  procedure pegacurva (color, x1, y1, x2, y2, x3, y3 : integer);
  procedure pegacurvas (color : byte; puntos : array of rpunto2; curvas :
    word);
  procedure pegacaja (color : byte; x1, y1, x2, y2 : integer);
  procedure pegacajallena (color : byte; x1, y1, x2, y2 : integer);
  procedure pegarueda (color : byte; x, y : integer; radiox, radioy : word);
  procedure pegaruedallena (color : byte; x, y : integer; radiox, radioy :
    word);

{/// Smbolo ///////////////////////////////////////////////////////////////}

  procedure ajustasimbolo;
  procedure expandesimbolo (porx, pory : integer);
  procedure pegasimbolos (color : byte; simbolos : pchar; x, y : integer);

{/// Cuadro ////////////////////////////////////////////////////////////////}

  procedure copiacuadro (x, y : integer; hacia : pointer; ancho, alto :
    word);
  procedure pegacuadro (desde : pointer; x, y : integer; ancho, alto : word);
  procedure pegacuadrohueco (desde : pointer; x, y : integer; ancho, alto :
    word; color : byte);
  procedure pegacuadrosolo (desde : pointer; x, y : integer; ancho, alto :
    word; color : byte);

{/// Paleta ////////////////////////////////////////////////////////////////}

  procedure acercacolor (color : byte; hasta : rmezcla3; cantidad : byte);

{/// Filtro ////////////////////////////////////////////////////////////////}

  procedure filtrapunto (x, y : integer);
  procedure filtralinea (x1, y1, x2, y2 : integer);
  procedure filtralineas (puntos : array of rpunto2; lineas : word);
  procedure filtracurva (x1, y1, x2, y2, x3, y3 : integer);
  procedure filtracurvas (puntos : array of rpunto2; curvas : word);
  procedure filtracaja (x1, y1, x2, y2 : integer);
  procedure filtracajallena (x1, y1, x2, y2 : integer);
  procedure filtrarueda (x, y : integer; radiox, radioy : word);
  procedure filtraruedallena (x, y : integer; radiox, radioy : word);
  procedure filtrasimbolos (simbolos : pchar; x, y : integer);

{/// Ratn /////////////////////////////////////////////////////////////////}

  procedure ajustaratonpagina;
  procedure ajustaratonplantilla;
  procedure centraraton;

implementation

  uses
    uvbs;

  const
    modo = $0101 {0103};

    anchopagina = 640 {800};
    altopagina = 480 {600};
    tamanopagina = anchopagina * altopagina;
    x1pagina = 0;
    y1pagina = 0;
    x2pagina = anchopagina - 1;
    y2pagina = altopagina - 1;
    xcentropagina = (anchopagina div 2) - 1;
    ycentropagina = (altopagina div 2) - 1;

    tamanoventana = 65536;
    ventana = 0;
    segmentoventana = $a000;

  var
    informemodo : rinformemodo;
    viejomodo : word;
    viejoexitproc : pointer;

    incrementoventana : word;
    lugarventana : word;

    paginas : word;
    paginaactiva : word;
    iniciopaginaactiva : longint;
    paginavisible : word;

    x1plantilla,
    y1plantilla,
    x2plantilla,
    y2plantilla : integer;

    viejallamadaraton : word;
    viejarutinaraton : pointer;

{/// Base //////////////////////////////////////////////////////////////////}

  function copiapuntobase (punto : longint) : byte;

    var
      lugar : word;

  begin
    lugar := word (punto div tamanoventana) * incrementoventana;

    if lugar <> lugarventana then
    begin
      usvbe. ponelugarventanadirecta (ventana, lugar);
      lugarventana := lugar;
    end;

    copiapuntobase := mem [segmentoventana:word (punto mod tamanoventana)];
  end;

  procedure pegapuntobase (color : byte; punto : longint);

    var
      lugar : word;

  begin
    lugar := word (punto div tamanoventana) * incrementoventana;

    if lugar <> lugarventana then
    begin
      usvbe. ponelugarventanadirecta (ventana, lugar);
      lugarventana := lugar;
    end;

    mem [segmentoventana:word (punto mod tamanoventana)] := color;
  end;

  procedure llenaventana (color : byte); assembler;
  asm
{Ajusta el color.}
    mov al, color
    mov ah, al
    mov bx, ax
    db 66h
    shl ax, 16
    mov ax, bx

{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    xor di, di

{Transfiere.}
    mov cx, 16384
    db 66h
    rep stosw
  end;

  procedure llenaparteventana (color : byte; desde, cantidad : word);
    assembler;
  asm
{Ajusta el color.}
    mov al, color
    mov ah, al
    mov bx, ax
    db 66h
    shl ax, 16
    mov ax, bx

{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    mov di, desde

{Transfiere.}
    mov cx, cantidad
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep stosw
    mov cx, bx
    rep stosb
  end;

  procedure llenapuntobase (color : byte; punto, cantidad : longint);

    var
      lugar,
      campo : word;
      parte : longint;

  begin
    lugar := word (punto div tamanoventana) * incrementoventana;

    if lugar <> lugarventana then
    begin
      usvbe. ponelugarventanadirecta (ventana, lugar);
      lugarventana := lugar;
    end;

    while cantidad > 0 do
    begin
      campo := word (punto mod tamanoventana);

      if cantidad + campo <= tamanoventana then
      begin  {Cabe en la ventana.}

        if cantidad = tamanoventana then
        begin
          llenaventana (color);
        end else
        begin
          llenaparteventana (color, campo, word (cantidad));
        end;

        cantidad := 0;
      end else
      begin  {Hay que partirlo.}
        parte := tamanoventana - campo;

        if parte = tamanoventana then
        begin
          llenaventana (color);
          inc (punto, tamanoventana);
          dec (cantidad, tamanoventana);
        end else
        begin
          llenaparteventana (color, campo, word (parte));
          inc (punto, parte);
          dec (cantidad, parte);
        end;

        inc (lugarventana, incrementoventana);
        usvbe. ponelugarventanadirecta (ventana, lugarventana);
      end;

    end;

  end;

  function copiaventana (hacia : pointer) : pointer; assembler;

    label
      salto, fin;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el destino.}
    les di, hacia

{Ajusta la fuente.}
    push ds
    mov si, segmentoventana
    mov ds, si
    xor si, si

{Salta si el destino apunta al inicio de su segmento, que es el caso cuando
se puede transferir una vez. Si no apunta transfiere dos veces.}
    test di, di
    jz salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    mov bx, ax
    sub ax, di
    inc ax
    mov parte1, ax
    sub bx, ax
    inc bx
    mov parte2, bx

{Transfiere la primera parte.}
    mov cx, parte1
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb

{Transfiere la segunda parte.}
    mov cx, parte2
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb
    jmp fin

    salto:

{Transfiere una vez.}
    mov cx, 16384
    db 66h
    rep movsw
{    jmp fin}

    fin:

{Regresa la direccin final del destino.}
    mov dx, es
    mov ax, di

    pop ds
  end;

  function copiaparteventana (desde : word; hacia : pointer; cantidad :
    word) : pointer; assembler;

    label
      salto, fin;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el destino.}
    les di, hacia

    mov cx, cantidad

{Ajusta la fuente.}
    push ds
    mov si, segmentoventana
    mov ds, si
    mov si, desde

{Si el campo del destino es menor o igual al mayor campo desde donde cabe
la transferencia, entonces hace una transferencia. Si no, hace dos
transferencias.}
    mov ax, 0ffffh
    sub ax, cx
    inc ax
    cmp di, ax
    jbe salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    sub ax, di
    inc ax
    mov parte1, ax
    sub cx, ax
    inc cx
    mov parte2, cx

{Transfiere la primera parte.}
    mov cx, parte1
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb

{Transfiere la segunda parte.}
    mov cx, parte2
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb
    jmp fin

    salto:

{Transfiere una vez.}
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb
{    jmp fin}

    fin:

{Regresa la direccin final del destino.}
    mov dx, es
    mov ax, di

    pop ds
  end;

  procedure copiapinturabase (punto : longint; hacia : pointer; cantidad :
    longint);

    var
      lugar,
      campo : word;
      parte : longint;

  begin
    lugar := word (punto div tamanoventana) * incrementoventana;

    if lugar <> lugarventana then
    begin
      usvbe. ponelugarventanadirecta (ventana, lugar);
      lugarventana := lugar;
    end;

    while cantidad > 0 do
    begin
      campo := word (punto mod tamanoventana);

      if cantidad + campo <= tamanoventana then
      begin  {Cabe en la ventana.}

        if cantidad = tamanoventana then
        begin
          copiaventana (hacia);
        end else
        begin
          copiaparteventana (campo, hacia, word (cantidad));
        end;

        cantidad := 0;
      end else
      begin  {Hay que partirlo.}
        parte := tamanoventana - campo;

        if parte = tamanoventana then
        begin
          hacia := copiaventana (hacia);
          inc (punto, tamanoventana);
          dec (cantidad, tamanoventana);
        end else
        begin
          hacia := copiaparteventana (campo, hacia, word (parte));
          inc (punto, parte);
          dec (cantidad, parte);
        end;

        inc (lugarventana, incrementoventana);
        usvbe. ponelugarventanadirecta (ventana, lugarventana);
      end;

    end;

  end;

  function pegaventana (desde : pointer) : pointer; assembler;

    label
      salto, fin;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    xor di, di

{Ajusta la fuente.}
    push ds
    lds si, desde

{Salta si la fuente apunta al inicio de su segmento, que es el caso cuando
se puede transferir una vez. Si no apunta al inicio transfiere dos veces.}
    test si, si
    jz salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    mov bx, ax
    sub ax, si
    inc ax
    mov parte1, ax
    sub bx, ax
    inc bx
    mov parte2, bx

{Transfiere la primera parte.}
    mov cx, parte1
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb

{Transfiere la segunda parte.}
    mov cx, parte2
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb
    jmp fin

    salto:

{Transfiere una vez.}
    mov cx, 16384
    db 66h
    rep movsw
{    jmp fin}

    fin:

{Regresa la direccin final de la fuente.}
    mov dx, ds
    mov ax, si

    pop ds
  end;

  function pegaparteventana (desde : pointer; hacia, cantidad : word) :
    pointer; assembler;

    label
      salto, fin;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    mov di, hacia

    mov cx, cantidad

{Ajusta la fuente.}
    push ds
    lds si, desde

{Si el campo de la fuente es menor o igual al mayor campo desde donde cabe
la transferencia, entonces hace una transferencia. Si no, hace dos
transferencias.}
    mov ax, 0ffffh
    sub ax, cx
    inc ax
    cmp si, ax
    jbe salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    sub ax, si
    inc ax
    mov parte1, ax
    sub cx, ax
    inc cx
    mov parte2, cx

{Transfiere la primera parte.}
    mov cx, parte1
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb

{Transfiere la segunda parte.}
    mov cx, parte2
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb
    jmp fin

    salto:

{Transfiere una vez.}
    mov bx, cx
    and bx, 3
    shr cx, 2
    db 66h
    rep movsw
    mov cx, bx
    rep movsb
{    jmp fin}

    fin:

{Regresa la direccin final de la fuente.}
    mov dx, ds
    mov ax, si

    pop ds
  end;

  procedure pegapinturabase (desde : pointer; punto, cantidad : longint);

    var
      lugar,
      campo : word;
      parte : longint;

  begin
    lugar := word (punto div tamanoventana) * incrementoventana;

    if lugar <> lugarventana then
    begin
      usvbe. ponelugarventanadirecta (ventana, lugar);
      lugarventana := lugar;
    end;

    while cantidad > 0 do
    begin
      campo := word (punto mod tamanoventana);

      if cantidad + campo <= tamanoventana then
      begin  {Cabe en la ventana.}

        if cantidad = tamanoventana then
        begin
          pegaventana (desde);
        end else
        begin
          pegaparteventana (desde, campo, word (cantidad));
        end;

        cantidad := 0;
      end else
      begin  {Hay que partirlo.}
        parte := tamanoventana - campo;

        if parte = tamanoventana then
        begin
          desde := pegaventana (desde);
          inc (punto, tamanoventana);
          dec (cantidad, tamanoventana);
        end else
        begin
          desde := pegaparteventana (desde, campo, word (parte));
          inc (punto, parte);
          dec (cantidad, parte);
        end;

        inc (lugarventana, incrementoventana);
        usvbe. ponelugarventanadirecta (ventana, lugarventana);
      end;

    end;

  end;

  function pegaventanahueco (desde : pointer; color : byte) : pointer;
    assembler;

    label
      salto, fin, ciclo1, ciclo2, ciclo3, siguiente1, siguiente2, siguiente3,
        fin1, fin2, fin3;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el excludo.}
    mov dl, color

{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    xor di, di

{Ajusta la fuente.}
    push ds
    lds si, desde

{Salta si la fuente apunta al inicio de su segmento, que es el caso cuando
se puede transferir una vez. Si no apunta al inicio transfiere dos veces.}
    test si, si
    jz salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    mov bx, ax
    sub ax, si
    inc ax
    mov parte1, ax
    sub bx, ax
    inc bx
    mov parte2, bx

{Transfiere la primera parte.}
    mov cx, parte1

    ciclo1:

    lodsb
    cmp al, dl
    je siguiente1
    stosb
    loop ciclo1
    jmp fin1

    siguiente1:

    inc di
    loop ciclo1

    fin1:

{Transfiere la segunda parte.}
    mov cx, parte2

    ciclo2:

    lodsb
    cmp al, dl
    je siguiente2
    stosb
    loop ciclo2
    jmp fin2

    siguiente2:

    inc di
    loop ciclo2

    fin2:

    jmp fin

    salto:

{Transfiere una vez.}

    ciclo3:

    lodsb
    cmp al, dl
    je siguiente3
    stosb
    test si, si
    jz ciclo3
    jmp fin3

    siguiente3:

    inc di
    test si, si
    jz ciclo3

    fin3:

{    jmp fin}

    fin:

{Regresa la direccin final de la fuente.}
    mov dx, ds
    mov ax, si

    pop ds
  end;

  function pegaparteventanahueco (desde : pointer; hacia, cantidad : word;
    color : byte) : pointer; assembler;

    label
      salto, fin, ciclo1, ciclo2, ciclo3, siguiente1, siguiente2, siguiente3,
        fin1, fin2, fin3;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el excludo.}
    mov dl, color

{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    mov di, hacia

    mov cx, cantidad

{Ajusta la fuente.}
    push ds
    lds si, desde

{Si el campo de la fuente es menor o igual al mayor campo desde donde cabe
la transferencia, entonces hace una transferencia. Si no, hace dos
transferencias.}
    mov ax, 0ffffh
    sub ax, cx
    inc ax
    cmp si, ax
    jbe salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    sub ax, si
    inc ax
    mov parte1, ax
    sub cx, ax
    inc cx
    mov parte2, cx

{Transfiere la primera parte.}
    mov cx, parte1

    ciclo1:

    lodsb
    cmp al, dl
    je siguiente1
    stosb
    loop ciclo1
    jmp fin1

    siguiente1:

    inc di
    loop ciclo1

    fin1:

{Transfiere la segunda parte.}
    mov cx, parte2

    ciclo2:

    lodsb
    cmp al, dl
    je siguiente2
    stosb
    loop ciclo2
    jmp fin2

    siguiente2:

    inc di
    loop ciclo2

    fin2:

    jmp fin

    salto:

{Transfiere una vez.}
    mov bx, cx

    ciclo3:

    lodsb
    cmp al, dl
    je siguiente3
    stosb
    loop ciclo3
    jmp fin3

    siguiente3:

    inc di
    loop ciclo3

    fin3:

{    jmp fin}

    fin:

{Regresa la direccin final de la fuente.}
    mov dx, ds
    mov ax, si

    pop ds
  end;

  procedure pegapinturahuecobase (desde : pointer; punto, cantidad : longint;
    color : byte);

    var
      lugar,
      campo : word;
      parte : longint;

  begin
    lugar := word (punto div tamanoventana) * incrementoventana;

    if lugar <> lugarventana then
    begin
      usvbe. ponelugarventanadirecta (ventana, lugar);
      lugarventana := lugar;
    end;

    while cantidad > 0 do
    begin
      campo := word (punto mod tamanoventana);

      if cantidad + campo <= tamanoventana then
      begin  {Cabe en la ventana.}

        if cantidad = tamanoventana then
        begin
          pegaventanahueco (desde, color);
        end else
        begin
          pegaparteventanahueco (desde, campo, word (cantidad), color);
        end;

        cantidad := 0;
      end else
      begin  {Hay que partirlo.}
        parte := tamanoventana - campo;

        if parte = tamanoventana then
        begin
          desde := pegaventanahueco (desde, color);
          inc (punto, tamanoventana);
          dec (cantidad, tamanoventana);
        end else
        begin
          desde := pegaparteventanahueco (desde, campo, word (parte), color);
          inc (punto, parte);
          dec (cantidad, parte);
        end;

        inc (lugarventana, incrementoventana);
        usvbe. ponelugarventanadirecta (ventana, lugarventana);
      end;

    end;

  end;

  function pegaventanasolo (desde : pointer; color : byte) : pointer;
    assembler;

    label
      salto, fin, ciclo1, ciclo2, ciclo3, siguiente1, siguiente2, siguiente3,
        fin1, fin2, fin3;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el excluyente.}
    mov dl, color

{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    xor di, di

{Ajusta la fuente.}
    push ds
    lds si, desde

{Salta si la fuente apunta al inicio de su segmento, que es el caso cuando
se puede transferir una vez. Si no apunta al inicio transfiere dos veces.}
    test si, si
    jz salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    mov bx, ax
    sub ax, si
    inc ax
    mov parte1, ax
    sub bx, ax
    inc bx
    mov parte2, bx

{Transfiere la primera parte.}
    mov cx, parte1

    ciclo1:

    lodsb
    cmp al, dl
    jne siguiente1
    stosb
    loop ciclo1
    jmp fin1

    siguiente1:

    inc di
    loop ciclo1

    fin1:

{Transfiere la segunda parte.}
    mov cx, parte2

    ciclo2:

    lodsb
    cmp al, dl
    jne siguiente2
    stosb
    loop ciclo2
    jmp fin2

    siguiente2:

    inc di
    loop ciclo2

    fin2:

    jmp fin

    salto:

{Transfiere una vez.}

    ciclo3:

    lodsb
    cmp al, dl
    jne siguiente3
    stosb
    test si, si
    jz ciclo3
    jmp fin3

    siguiente3:

    inc di
    test si, si
    jz ciclo3

    fin3:

{    jmp fin}

    fin:

{Regresa la direccin final de la fuente.}
    mov dx, ds
    mov ax, si

    pop ds
  end;

  function pegaparteventanasolo (desde : pointer; hacia, cantidad : word;
    color : byte) : pointer; assembler;

    label
      salto, fin, ciclo1, ciclo2, ciclo3, siguiente1, siguiente2, siguiente3,
        fin1, fin2, fin3;

    var
      parte1,
      parte2 : word;

  asm
{Ajusta el excluyente.}
    mov dl, color

{Ajusta el destino.}
    mov di, segmentoventana
    mov es, di
    mov di, hacia

    mov cx, cantidad

{Ajusta la fuente.}
    push ds
    lds si, desde

{Si el campo de la fuente es menor o igual al mayor campo desde donde cabe
la transferencia, entonces hace una transferencia. Si no, hace dos
transferencias.}
    mov ax, 0ffffh
    sub ax, cx
    inc ax
    cmp si, ax
    jbe salto

{Calcula el tamao de las dos partes y las salva.}
    mov ax, 0ffffh
    sub ax, si
    inc ax
    mov parte1, ax
    sub cx, ax
    inc cx
    mov parte2, cx

{Transfiere la primera parte.}
    mov cx, parte1

    ciclo1:

    lodsb
    cmp al, dl
    jne siguiente1
    stosb
    loop ciclo1
    jmp fin1

    siguiente1:

    inc di
    loop ciclo1

    fin1:

{Transfiere la segunda parte.}
    mov cx, parte2

    ciclo2:

    lodsb
    cmp al, dl
    jne siguiente2
    stosb
    loop ciclo2
    jmp fin2

    siguiente2:

    inc di
    loop ciclo2

    fin2:

    jmp fin

    salto:

{Transfiere una vez.}
    mov bx, cx

    ciclo3:

    lodsb
    cmp al, dl
    jne siguiente3
    stosb
    loop ciclo3
    jmp fin3

    siguiente3:

    inc di
    loop ciclo3

    fin3:

{    jmp fin}

    fin:

{Regresa la direccin final de la fuente.}
    mov dx, ds
    mov ax, si

    pop ds
  end;

  procedure pegapinturasolobase (desde : pointer; punto, cantidad : longint;
    color : byte);

    var
      lugar,
      campo : word;
      parte : longint;

  begin
    lugar := word (punto div tamanoventana) * incrementoventana;

    if lugar <> lugarventana then
    begin
      usvbe. ponelugarventanadirecta (ventana, lugar);
      lugarventana := lugar;
    end;

    while cantidad > 0 do
    begin
      campo := word (punto mod tamanoventana);

      if cantidad + campo <= tamanoventana then
      begin  {Cabe en la ventana.}

        if cantidad = tamanoventana then
        begin
          pegaventanasolo (desde, color);
        end else
        begin
          pegaparteventanasolo (desde, campo, word (cantidad), color);
        end;

        cantidad := 0;
      end else
      begin  {Hay que partirlo.}
        parte := tamanoventana - campo;

        if parte = tamanoventana then
        begin
          desde := pegaventanasolo (desde, color);
          inc (punto, tamanoventana);
          dec (cantidad, tamanoventana);
        end else
        begin
          desde := pegaparteventanasolo (desde, campo, word (parte), color);
          inc (punto, parte);
          dec (cantidad, parte);
        end;

        inc (lugarventana, incrementoventana);
        usvbe. ponelugarventanadirecta (ventana, lugarventana);
      end;

    end;

  end;

{/// Pgina ////////////////////////////////////////////////////////////////}

  function tomapaginas : word;
  begin
    tomapaginas := paginas;
  end;

  procedure ponepaginaactiva (pagina : word);
  begin

    if pagina < paginas then
    begin
      paginaactiva := pagina;
      iniciopaginaactiva := longint (pagina) * tamanopagina;
    end;

  end;

  function tomapaginaactiva : word;
  begin
    tomapaginaactiva := paginaactiva;
  end;

  procedure ponepaginavisible (pagina : word);

    var
      inicio : longint;

  begin

    if pagina < paginas then
    begin
      inicio := longint (pagina) * tamanopagina;
      poneiniciopantalla (word (inicio div anchopagina), word (inicio mod
        anchopagina));
      paginavisible := pagina;
    end;

  end;

  function tomapaginavisible : word;
  begin
    tomapaginavisible := paginavisible;
  end;

  procedure regresapaginaactiva;
  begin

    if paginaactiva = 0 then
    begin
      paginaactiva := paginas - 1;
    end else
    begin
      dec (paginaactiva);
    end;

    iniciopaginaactiva := longint (paginaactiva) * tamanopagina;
  end;

  procedure avanzapaginaactiva;
  begin

    if paginaactiva >= paginas - 1 then
    begin
      paginaactiva := 0;
    end else
    begin
      inc (paginaactiva);
    end;

    iniciopaginaactiva := longint (paginaactiva) * tamanopagina;
  end;

  procedure regresapaginavisible;

    var
      inicio : longint;

  begin

    if paginavisible = 0 then
    begin
      paginavisible := paginas - 1;
    end else
    begin
      dec (paginavisible);
    end;

    inicio := longint (paginavisible) * tamanopagina;
    poneiniciopantalla (word (inicio div anchopagina), word (inicio mod
      anchopagina));
  end;

  procedure avanzapaginavisible;

    var
      inicio : longint;

  begin

    if paginavisible >= paginas - 1 then
    begin
      paginavisible := 0;
    end else
    begin
      inc (paginavisible);
    end;

    inicio := longint (paginavisible) * tamanopagina;
    poneiniciopantalla (word (inicio div anchopagina), word (inicio mod
      anchopagina));
  end;

  procedure metepagina;
  begin
    paginas := byte ((longint (informe. memoria) * unidadmemoria) div
      tamanopagina);
    paginaactiva := 0;
    iniciopaginaactiva := 0;
    paginavisible := 0;
  end;

{/// Plantilla /////////////////////////////////////////////////////////////}

  procedure ajustaplantilla;
  begin
    x1plantilla := x1pagina;
    y1plantilla := y1pagina;
    x2plantilla := x2pagina;
    y2plantilla := y2pagina;
  end;

  procedure tomaplantilla (var x1, y1, x2, y2 : integer);
  begin
    x1 := x1plantilla;
    y1 := y1plantilla;
    x2 := x2plantilla;
    y2 := y2plantilla;
  end;

  function anchoplantilla : word;
  begin
    anchoplantilla := x2plantilla - x1plantilla + 1;
  end;

  function altoplantilla : word;
  begin
    altoplantilla := y2plantilla - y1plantilla + 1;
  end;

  procedure poneplantilla (x1, y1, x2, y2 : integer);

    var
      cambio : integer;

  begin
{Suma.}
    x1plantilla := x1;
    y1plantilla := y1;
    x2plantilla := x2;
    y2plantilla := y2;

{Ajusta.}
    if x1plantilla > x2plantilla then
    begin
      cambio := x1plantilla;
      x1plantilla := x2plantilla;
      x2plantilla := cambio;
    end;

    if y1plantilla > y2plantilla then
    begin
      cambio := y1plantilla;
      y1plantilla := y2plantilla;
      y2plantilla := cambio;
    end;

    if x1plantilla < x1pagina then x1plantilla := x1pagina;
    if y1plantilla < y1pagina then y1plantilla := y1pagina;
    if x2plantilla > x2pagina then x2plantilla := x2pagina;
    if y2plantilla > y2pagina then y2plantilla := y2pagina;
  end;

  procedure mueveplantilla (masx, masy : integer);

    var
      cambio : integer;

  begin
{Mueve.}
    inc (x1plantilla, masx);
    inc (y1plantilla, masy);
    inc (x2plantilla, masx);
    inc (y2plantilla, masy);

{Ajusta.}
    if x1plantilla > x2plantilla then
    begin
      cambio := x1plantilla;
      x1plantilla := x2plantilla;
      x2plantilla := cambio;
    end;

    if y1plantilla > y2plantilla then
    begin
      cambio := y1plantilla;
      y1plantilla := y2plantilla;
      y2plantilla := cambio;
    end;

    if x1plantilla < x1pagina then x1plantilla := x1pagina;
    if y1plantilla < y1pagina then y1plantilla := y1pagina;
    if x2plantilla > x2pagina then x2plantilla := x2pagina;
    if y2plantilla > y2pagina then y2plantilla := y2pagina;
  end;

  procedure creceplantilla (masx, masy : integer);

    var
      cambio : integer;

  begin
{Crece.}
    dec (x1plantilla, masx);
    dec (y1plantilla, masy);
    inc (x2plantilla, masx);
    inc (y2plantilla, masy);

{Ajusta.}
    if x1plantilla > x2plantilla then
    begin
      cambio := x1plantilla;
      x1plantilla := x2plantilla;
      x2plantilla := cambio;
    end;

    if y1plantilla > y2plantilla then
    begin
      cambio := y1plantilla;
      y1plantilla := y2plantilla;
      y2plantilla := cambio;
    end;

    if x1plantilla < x1pagina then x1plantilla := x1pagina;
    if y1plantilla < y1pagina then y1plantilla := y1pagina;
    if x2plantilla > x2pagina then x2plantilla := x2pagina;
    if y2plantilla > y2pagina then y2plantilla := y2pagina;
  end;

  function puntodentroplantilla (x, y : integer) : boolean;
  begin
    puntodentroplantilla := (x >= x1plantilla) and (y >= y1plantilla) and
      (x <= x2plantilla) and (y <= y2plantilla);
  end;

  function cajadentroplantilla (x1, y1, x2, y2 : integer) : boolean;

    var
      cambio : integer;

  begin

{Ajusta.}
    if x1 > x2 then
    begin
      cambio := x1;
      x1 := x2;
      x2 := cambio;
    end;

    if y1 > y2 then
    begin
      cambio := y1;
      y1 := y2;
      y2 := cambio;
    end;

{Regresa true si toda la caja o parte de ella est dentro de la plantilla.}
    cajadentroplantilla := (x1 <= x2plantilla) and (y1 <= y2plantilla) and
      (x2 >= x1plantilla) and (y2 >= y1plantilla);
  end;

  function cuadrodentroplantilla (x, y : integer; ancho, alto : word) :
    boolean;
  begin
{Regresa true si todo el cuadro o parte de l est dentro de la plantilla.}
    cuadrodentroplantilla := (x <= x2plantilla) and (y <= y2plantilla) and
      (x + ancho > x1plantilla) and (y + alto > y1plantilla);
  end;

  procedure meteplantilla;
  begin
    ajustaplantilla;
  end;

{/// Pintura ///////////////////////////////////////////////////////////////}

  function copiapunto (x, y : integer) : byte;
  begin

    if puntodentroplantilla (x, y) then
    begin
      copiapunto := copiapuntobase (longint (y) * anchopagina + x +
        iniciopaginaactiva);
    end else
    begin
      copiapunto := 0;
    end;

  end;

  procedure pegapunto (color : byte; x, y : integer);
  begin

    if puntodentroplantilla (x, y) then
    begin
      pegapuntobase (color, longint (y) * anchopagina + x +
        iniciopaginaactiva);
    end;

  end;

{Basada en la rutina de Menno Victor van der star}
  procedure pegalinea (color : byte; x1, y1, x2, y2 : integer);

    var
      d,
      ax,
      ay,
      sx,
      sy,
      dx,
      dy : Integer;

  begin
    dx := x2 - x1;
    ax := abs (dx) shl 1;
    if dx < 0 then sx := -1 else sx := 1;
    dy := y2 - y1;
    ay := abs (dy) shl 1;
    if dy < 0 then sy := -1 else sy := 1;
    pegapunto (color, x1, y1);

    if ax > ay then
    begin
      d := ay - (ax shr 1);

      while x1 <> x2 do
      begin

        if d >= 0 then
        begin
          inc (y1, sy);
          dec (d, ax);
        end;

        inc (x1, sx);
        inc (d, ay);
        pegapunto (color, x1, y1);
      end;

    end else
    begin
      d := ax - (ay shr 1);

      while y1 <> y2 do
      begin

        if d >= 0 then
        begin
          inc (x1, sx);
          dec (d, ay);
        end;

        inc (y1, sy);
        inc (d, ax);
        pegapunto (color, x1, y1);
      end;

    end;

  end;

  procedure pegalineas (color : byte; puntos : array of rpunto2; lineas :
    word);

    var
      indice : word;

  begin
    indice := 0;

    for lineas := lineas downto 1 do
    begin
      pegalinea (color, puntos [indice]. x, puntos [indice]. y,
        puntos [indice + 1]. x, puntos [indice + 1]. y);
      inc (indice);
    end;

  end;

{Basada en la rutina de Wil Barath}
  procedure pegacurva (color, x1, y1, x2, y2, x3, y3 : integer);

    var
      sx,
      sy,
      dx1,
      dx2,
      dy1,
      dy2,
      sdx,
      sdy,
      l : longint;
      deep,
      deep2,
      ox,
      oy : integer;

  begin
    dx1 := (x2 - x1);
    dx2 := (x3 - x2);
    dy1 := (y2 - y1);
    dy2 := (y3 - y2);
    deep := 32;
    deep2 := deep * deep;
    sdx := dx1 * pred (deep * 2) - dx2;
    sdy := dy1 * pred (deep * 2) - dy2;
    inc (dx1, dx1);
    inc (dx2, dx2);
    inc (dy1, dy1);
    inc (dy2, dy2);
    sx := 0;
    sy := 0;
    x2 := x1;
    y2 := y1;

    for l := 0 to deep do
    begin
      ox := x2;
      oy := y2;
      x2 := x1 + sx div deep2;
      y2 := y1 + sy div deep2;
      pegalinea (color, x2, y2, ox, oy);
      inc (sx, sdx);
      inc (sy, sdy);
      inc (sdx, dx2 - dx1);
      inc (sdy, dy2 - dy1);
    end;

    pegalinea (color, x2, y2, x3, y3);
  end;

  procedure pegacurvas (color : byte; puntos : array of rpunto2; curvas :
    word);

    var
      indice : word;

  begin
    indice := 0;

    for curvas := curvas downto 1 do
    begin
      pegacurva (color, puntos [indice]. x, puntos [indice]. y,
        puntos [indice + 1]. x, puntos [indice + 1]. y,
        puntos [indice + 2]. x, puntos [indice + 2]. y);
      inc (indice, 2);
    end;

  end;

  procedure pegacaja (color : byte; x1, y1, x2, y2 : integer);
  begin
    pegalinea (color, x1, y1, x2, y1);
    pegalinea (color, x2, y1, x2, y2);
    pegalinea (color, x2, y2, x1, y2);
    pegalinea (color, x1, y2, x1, y1);
  end;

  procedure pegacajallena (color : byte; x1, y1, x2, y2 : integer);

    var
      cambio : integer;
      punto,
      ancho : longint;

  begin

    if cuadrodentroplantilla (x1, y1, x2, y2) then
    begin

{Ajusta.}
      if x1 > x2 then
      begin
        cambio := x1;
        x1 := x2;
        x2 := cambio;
      end;

      if y1 > y2 then
      begin
        cambio := y1;
        y1 := y2;
        y2 := cambio;
      end;

      if x1 < x1plantilla then x1 := x1plantilla;
      if y1 < y1plantilla then y1 := y1plantilla;
      if x2 > x2plantilla then x2 := x2plantilla;
      if y2 > y2plantilla then y2 := y2plantilla;

      punto := longint (y1) * anchopagina + x1 + iniciopaginaactiva;
      ancho := x2 - x1 + 1;

      for y1 := y1 to y2 do
      begin
        llenapuntobase (color, punto, ancho);
        inc (punto, anchopagina);
      end;

    end;

  end;

{Basada en la rutina de Gerd, correo: gerd.platl@siemens.at}
  procedure pegarueda (color : byte; x, y : integer; radiox, radioy : word);

    var
      radio,
      x1,
      x2,
      y1,
      y2 : integer;
      aq,
      bq,
      dx,
      dy,
      r,
      rx,
      ry : longint;

  begin
    pegapunto (color, x + radiox, y);
    pegapunto (color, x - radiox, y);
    x1 := x - radiox;
    y1 := y;
    x2 := x + radiox;
    y2 := y;
    aq := longint (radiox) * radiox;
    bq := longint (radioy) * radioy;
    dx := aq shl 1;
    dy := bq shl 1;
    r := radiox * bq;
    rx := r shl 1;
    ry := 0;
    radio := radiox;

    while radio > 0 do
    begin

      if r > 0 then
      begin
        inc (y1);
        dec (y2);
        inc (ry, dx);
        dec (r, ry);
      end;

      if r <= 0 then
      begin
        dec (radio);
        inc (x1);
        dec (x2);
        dec (rx, dy);
        inc (r, rx);
      end;

      pegapunto (color, x1, y1);
      pegapunto (color, x1, y2);
      pegapunto (color, x2, y1);
      pegapunto (color, x2, y2);
    end;

  end;

  procedure pegaruedallena (color : byte; x, y : integer; radiox, radioy :
    word);

    var
      radio,
      x1,
      x2,
      y1,
      y2 : integer;
      aq,
      bq,
      dx,
      dy,
      r,
      rx,
      ry : longint;

  begin
    pegalinea (color, x - radiox, y, x + radiox, y);
    x1 := x - radiox;
    y1 := y;
    x2 := x + radiox;
    y2 := y;
    aq := longint (radiox) * radiox;
    bq := longint (radioy) * radioy;
    dx := aq shl 1;
    dy := bq shl 1;
    r := radiox * bq;
    rx := r shl 1;
    ry := 0;
    radio := radiox;

    while radio > 0 do
    begin

      if r > 0 then
      begin
        inc (y1);
        dec (y2);
        inc (ry, dx);
        dec (r, ry);
      end;

      if r <= 0 then
      begin
        dec (radio);
        inc (x1);
        dec (x2);
        dec (rx, dy);
        inc (r, rx);
      end;

      pegalinea (color, x1, y1, x2, y1);
      pegalinea (color, x1, y2, x2, y2);
    end;

  end;

{/// Smbolo ///////////////////////////////////////////////////////////////}

  procedure ajustasimbolo;
  begin
    masx1simbolo := 1;
    masy1simbolo := 0;
    masx2simbolo := -7;
    masy2simbolo := 1;
    masx3simbolo := 8;
    masy3simbolo := -8;
  end;

  procedure expandesimbolo (porx, pory : integer);
  begin
    masx1simbolo := masx1simbolo * porx;
    masy1simbolo := masy1simbolo * pory;
    masx2simbolo := masx2simbolo * porx;
    masy2simbolo := masy2simbolo * pory;
    masx3simbolo := masx3simbolo * porx;
    masy3simbolo := masy3simbolo * pory;
  end;

  procedure pegasimbolos (color : byte; simbolos : pchar; x, y : integer);

    var
      indicesimbolo : word;
      indicetabla : pointer;
      indicelinea,
      linea : byte;

  begin
    indicesimbolo := 0;

    while simbolos [indicesimbolo] <> #0 do
    begin
      indicetabla := ptr (seg (fundicion^), ofs (fundicion^) + word
        (simbolos [indicesimbolo]) * 8);

      for indicelinea := 0 to 7 do
      begin
        linea := byte (indicetabla^);
        if linea and 128 <> 0 then pegapunto (color, x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 64 <> 0 then pegapunto (color, x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 32 <> 0 then pegapunto (color, x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 16 <> 0 then pegapunto (color, x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 8 <> 0 then pegapunto (color, x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 4 <> 0 then pegapunto (color, x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 2 <> 0 then pegapunto (color, x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 1 <> 0 then pegapunto (color, x, y);
        inc (x, masx2simbolo);
        inc (y, masy2simbolo);
        inc (word (indicetabla));
      end;

      inc (x, masx3simbolo);
      inc (y, masy3simbolo);
      inc (indicesimbolo);
    end;

  end;

  procedure metesimbolo;
  begin
    fundicion := uvbs. tomafundicion;
    ajustasimbolo;
  end;

{/// Cuadro ////////////////////////////////////////////////////////////////}

  procedure copiacuadro (x, y : integer; hacia : pointer; ancho, alto :
    word);

    var
      incremento : longint;
      ancho1,
      alto1 : word;
      punto : longint;

  begin

    if cuadrodentroplantilla (x, y, ancho, alto) then
    begin

{Corta los lados que estn fuera de la plantilla.}
      if x >= x1plantilla then
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la derecha. Reduce el ancho a la distancia entre x y x2plantilla.}
          ancho1 := x2plantilla - x + 1;
        end else
        begin
{No corta. Mantiene el ancho.}
          ancho1 := ancho;
        end;

      end else
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la izquierda y a la derecha.}
          ancho1 := anchoplantilla;
        end else
        begin
{Corta a la izquierda.}
          ancho1 := ancho - (x1plantilla - x);
        end;

        hacia := sumapunterocerca (hacia, x1plantilla - x);
        x := x1plantilla;
      end;

      if y >= y1plantilla then
      begin

        if y + alto > y2plantilla then
        begin
{Corta abajo.}
          alto1 := y2plantilla - y + 1;
        end else
        begin
{No corta.}
          alto1 := alto;
        end;

      end else
      begin

        if y + alto > y2plantilla then
        begin
{Corta arriba y abajo.}
          alto1 := altoplantilla;
        end else
        begin
{Corta arriba.}
          alto1 := alto - (y1plantilla - y);
        end;

        incremento := longint (y1plantilla - y) * ancho;
        hacia := sumapunterolejos (hacia, incremento div 16,
          incremento mod 16);
        y := y1plantilla;
      end;

      punto := longint (y) * anchopagina + x + iniciopaginaactiva;

      for alto1 := alto1 downto 1 do
      begin
        copiapinturabase (punto, hacia, ancho1);
        inc (punto, anchopagina);
        hacia := sumapunterocerca (hacia, ancho);
      end;

    end;

  end;

  procedure pegacuadro (desde : pointer; x, y : integer; ancho, alto : word);

    var
      incremento : longint;
      ancho1,
      alto1 : word;
      punto : longint;

  begin

    if cuadrodentroplantilla (x, y, ancho, alto) then
    begin

{Corta los lados que estn fuera de la plantilla.}
      if x >= x1plantilla then
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la derecha. Reduce el ancho a la distancia entre x y x2plantilla.}
          ancho1 := x2plantilla - x + 1;
        end else
        begin
{No corta. Mantiene el ancho.}
          ancho1 := ancho;
        end;

      end else
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la izquierda y a la derecha.}
          ancho1 := anchoplantilla;
        end else
        begin
{Corta a la izquierda.}
          ancho1 := ancho - (x1plantilla - x);
        end;

        desde := sumapunterocerca (desde, x1plantilla - x);
        x := x1plantilla;
      end;

      if y >= y1plantilla then
      begin

        if y + alto > y2plantilla then
        begin
{Corta abajo.}
          alto1 := y2plantilla - y + 1;
        end else
        begin
{No corta.}
          alto1 := alto;
        end;

      end else
      begin

        if y + alto > y2plantilla then
        begin
{Corta arriba y abajo.}
          alto1 := altoplantilla;
        end else
        begin
{Corta arriba.}
          alto1 := alto - (y1plantilla - y);
        end;

        incremento := longint (y1plantilla - y) * ancho;
        desde := sumapunterolejos (desde, incremento div 16,
          incremento mod 16);
        y := y1plantilla;
      end;

      punto := longint (y) * anchopagina + x + iniciopaginaactiva;

      for alto1 := alto1 downto 1 do
      begin
        pegapinturabase (desde, punto, ancho1);
        inc (punto, anchopagina);
        desde := sumapunterocerca (desde, ancho);
      end;

    end;

  end;

  procedure pegacuadrohueco (desde : pointer; x, y : integer; ancho, alto :
    word; color : byte);

    var
      incremento : longint;
      ancho1,
      alto1 : word;
      punto : longint;

  begin

    if cuadrodentroplantilla (x, y, ancho, alto) then
    begin

{Corta los lados que estn fuera de la plantilla.}
      if x >= x1plantilla then
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la derecha. Reduce el ancho a la distancia entre x y x2plantilla.}
          ancho1 := x2plantilla - x + 1;
        end else
        begin
{No corta. Mantiene el ancho.}
          ancho1 := ancho;
        end;

      end else
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la izquierda y a la derecha.}
          ancho1 := anchoplantilla;
        end else
        begin
{Corta a la izquierda.}
          ancho1 := ancho - (x1plantilla - x);
        end;

        desde := sumapunterocerca (desde, x1plantilla - x);
        x := x1plantilla;
      end;

      if y >= y1plantilla then
      begin

        if y + alto > y2plantilla then
        begin
{Corta abajo.}
          alto1 := y2plantilla - y + 1;
        end else
        begin
{No corta.}
          alto1 := alto;
        end;

      end else
      begin

        if y + alto > y2plantilla then
        begin
{Corta arriba y abajo.}
          alto1 := altoplantilla;
        end else
        begin
{Corta arriba.}
          alto1 := alto - (y1plantilla - y);
        end;

        incremento := longint (y1plantilla - y) * ancho;
        desde := sumapunterolejos (desde, incremento div 16,
          incremento mod 16);
        y := y1plantilla;
      end;

      punto := longint (y) * anchopagina + x + iniciopaginaactiva;

      for alto1 := alto1 downto 1 do
      begin
        pegapinturahuecobase (desde, punto, ancho1, color);
        inc (punto, anchopagina);
        desde := sumapunterocerca (desde, ancho);
      end;

    end;

  end;

  procedure pegacuadrosolo (desde : pointer; x, y : integer; ancho, alto :
    word; color : byte);

    var
      incremento : longint;
      ancho1,
      alto1 : word;
      punto : longint;

  begin

    if cuadrodentroplantilla (x, y, ancho, alto) then
    begin

{Corta los lados que estn fuera de la plantilla.}
      if x >= x1plantilla then
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la derecha. Reduce el ancho a la distancia entre x y x2plantilla.}
          ancho1 := x2plantilla - x + 1;
        end else
        begin
{No corta. Mantiene el ancho.}
          ancho1 := ancho;
        end;

      end else
      begin

        if x + ancho > x2plantilla then
        begin
{Corta a la izquierda y a la derecha.}
          ancho1 := anchoplantilla;
        end else
        begin
{Corta a la izquierda.}
          ancho1 := ancho - (x1plantilla - x);
        end;

        desde := sumapunterocerca (desde, x1plantilla - x);
        x := x1plantilla;
      end;

      if y >= y1plantilla then
      begin

        if y + alto > y2plantilla then
        begin
{Corta abajo.}
          alto1 := y2plantilla - y + 1;
        end else
        begin
{No corta.}
          alto1 := alto;
        end;

      end else
      begin

        if y + alto > y2plantilla then
        begin
{Corta arriba y abajo.}
          alto1 := altoplantilla;
        end else
        begin
{Corta arriba.}
          alto1 := alto - (y1plantilla - y);
        end;

        incremento := longint (y1plantilla - y) * ancho;
        desde := sumapunterolejos (desde, incremento div 16,
          incremento mod 16);
        y := y1plantilla;
      end;

      punto := longint (y) * anchopagina + x + iniciopaginaactiva;

      for alto1 := alto1 downto 1 do
      begin
        pegapinturasolobase (desde, punto, ancho1, color);
        inc (punto, anchopagina);
        desde := sumapunterocerca (desde, ancho);
      end;

    end;

  end;

{/// Paleta ////////////////////////////////////////////////////////////////}

  procedure acercacolor (color : byte; hasta : rmezcla3; cantidad : byte);

    var
      mezcla : rmezcla3;

  begin
    uvga. copiamezcla (color, @mezcla, sizeof (mezcla));

    for cantidad := cantidad downto 1 do
    begin
      if mezcla. rojo > hasta. rojo then dec (mezcla. rojo);
      if mezcla. rojo < hasta. rojo then inc (mezcla. rojo);
      if mezcla. verde > hasta. verde then dec (mezcla. verde);
      if mezcla. verde < hasta. verde then inc (mezcla. verde);
      if mezcla. azul > hasta. azul then dec (mezcla. azul);
      if mezcla. azul < hasta. azul then inc (mezcla. azul);
    end;

    uvga. pegamezcla (@mezcla, color, sizeof (mezcla));
  end;

{/// Filtro ////////////////////////////////////////////////////////////////}

  procedure filtrapunto (x, y : integer);

    var
      resultado : integer;

  begin
{Multiplica los puntos por los coeficientes de la tabla y los suma.}
    resultado := 0;

    inc (resultado, filtro. coeficientes [0] * copiapunto (x - 2, y - 2));
    inc (resultado, filtro. coeficientes [1] * copiapunto (x - 1, y - 2));
    inc (resultado, filtro. coeficientes [2] * copiapunto (x, y - 2));
    inc (resultado, filtro. coeficientes [3] * copiapunto (x + 1, y - 2));
    inc (resultado, filtro. coeficientes [4] * copiapunto (x + 2, y - 2));

    inc (resultado, filtro. coeficientes [5] * copiapunto (x - 2, y - 1));
    inc (resultado, filtro. coeficientes [6] * copiapunto (x - 1, y - 1));
    inc (resultado, filtro. coeficientes [7] * copiapunto (x, y - 1));
    inc (resultado, filtro. coeficientes [8] * copiapunto (x + 1, y - 1));
    inc (resultado, filtro. coeficientes [9] * copiapunto (x + 2, y - 1));

    inc (resultado, filtro. coeficientes [10] * copiapunto (x - 2, y));
    inc (resultado, filtro. coeficientes [11] * copiapunto (x - 1, y));
    inc (resultado, filtro. coeficientes [12] * copiapunto (x, y));
    inc (resultado, filtro. coeficientes [13] * copiapunto (x + 1, y));
    inc (resultado, filtro. coeficientes [14] * copiapunto (x + 2, y));

    inc (resultado, filtro. coeficientes [15] * copiapunto (x - 2, y + 1));
    inc (resultado, filtro. coeficientes [16] * copiapunto (x - 1, y + 1));
    inc (resultado, filtro. coeficientes [17] * copiapunto (x, y + 1));
    inc (resultado, filtro. coeficientes [18] * copiapunto (x + 1, y + 1));
    inc (resultado, filtro. coeficientes [19] * copiapunto (x + 2, y + 1));

    inc (resultado, filtro. coeficientes [20] * copiapunto (x - 2, y + 2));
    inc (resultado, filtro. coeficientes [21] * copiapunto (x - 1, y + 2));
    inc (resultado, filtro. coeficientes [22] * copiapunto (x, y + 2));
    inc (resultado, filtro. coeficientes [23] * copiapunto (x + 1, y + 2));
    inc (resultado, filtro. coeficientes [24] * copiapunto (x + 2, y + 2));

{Divide entre el factor.}
    resultado := resultado div filtro. factor;

{Suma el incremento.}
    inc (resultado, filtro. incremento);

{Pega el punto.}
    pegapunto (byte (resultado), x, y);
  end;

  procedure filtralinea (x1, y1, x2, y2 : integer);

    var
      d,
      ax,
      ay,
      sx,
      sy,
      dx,
      dy : Integer;

  begin
    dx := x2 - x1;
    ax := abs (dx) shl 1;
    if dx < 0 then sx := -1 else sx := 1;
    dy := y2 - y1;
    ay := abs (dy) shl 1;
    if dy < 0 then sy := -1 else sy := 1;
    filtrapunto (x1, y1);

    if ax > ay then
    begin
      d := ay - (ax shr 1);

      while x1 <> x2 do
      begin

        if d >= 0 then
        begin
          inc (y1, sy);
          dec (d, ax);
        end;

        inc (x1, sx);
        inc (d, ay);
        filtrapunto (x1, y1);
      end;

    end else
    begin
      d := ax - (ay shr 1);

      while y1 <> y2 do
      begin

        if d >= 0 then
        begin
          inc (x1, sx);
          dec (d, ay);
        end;

        inc (y1, sy);
        inc (d, ax);
        filtrapunto (x1, y1);
      end;

    end;

  end;

  procedure filtralineas (puntos : array of rpunto2; lineas : word);

    var
      indice : word;

  begin
    indice := 0;

    for lineas := lineas downto 1 do
    begin
      filtralinea (puntos [indice]. x, puntos [indice]. y,
        puntos [indice + 1]. x, puntos [indice + 1]. y);
      inc (indice);
    end;

  end;

  procedure filtracurva (x1, y1, x2, y2, x3, y3 : integer);

    var
      sx,
      sy,
      dx1,
      dx2,
      dy1,
      dy2,
      sdx,
      sdy,
      l : longint;
      deep,
      deep2,
      ox,
      oy : integer;

  begin
    dx1 := (x2 - x1);
    dx2 := (x3 - x2);
    dy1 := (y2 - y1);
    dy2 := (y3 - y2);
    deep := 32;
    deep2 := deep * deep;
    sdx := dx1 * pred (deep * 2) - dx2;
    sdy := dy1 * pred (deep * 2) - dy2;
    inc (dx1, dx1);
    inc (dx2, dx2);
    inc (dy1, dy1);
    inc (dy2, dy2);
    sx := 0;
    sy := 0;
    x2 := x1;
    y2 := y1;

    for l := 0 to deep do
    begin
      ox := x2;
      oy := y2;
      x2 := x1 + sx div deep2;
      y2 := y1 + sy div deep2;
      filtralinea (x2, y2, ox, oy);
      inc (sx, sdx);
      inc (sy, sdy);
      inc (sdx, dx2 - dx1);
      inc (sdy, dy2 - dy1);
    end;

    filtralinea (x2, y2, x3, y3);
  end;

  procedure filtracurvas (puntos : array of rpunto2; curvas : word);

    var
      indice : word;

  begin
    indice := 0;

    for curvas := curvas downto 1 do
    begin
      filtracurva (puntos [indice]. x, puntos [indice]. y,
        puntos [indice + 1]. x, puntos [indice + 1]. y,
        puntos [indice + 2]. x, puntos [indice + 2]. y);
      inc (indice, 2);
    end;

  end;

  procedure filtracaja (x1, y1, x2, y2 : integer);
  begin
    filtralinea (x1, y1, x2, y1);
    filtralinea (x2, y1, x2, y2);
    filtralinea (x2, y2, x1, y2);
    filtralinea (x1, y2, x1, y1);
  end;

  procedure filtracajallena (x1, y1, x2, y2 : integer);
  begin

    if y1 <= y2 then
    begin
      for y1 := y1 to y2 do filtralinea (x1, y1, x2, y1);
    end else
    begin
      for y2 := y2 to y1 do filtralinea (x1, y2, x2, y2);
    end;

  end;

  procedure filtrarueda (x, y : integer; radiox, radioy : word);

    var
      radio,
      x1,
      x2,
      y1,
      y2 : integer;
      aq,
      bq,
      dx,
      dy,
      r,
      rx,
      ry : longint;

  begin
    filtrapunto (x + radiox, y);
    filtrapunto (x - radiox, y);
    x1 := x - radiox;
    y1 := y;
    x2 := x + radiox;
    y2 := y;
    aq := longint (radiox) * radiox;
    bq := longint (radioy) * radioy;
    dx := aq shl 1;
    dy := bq shl 1;
    r := radiox * bq;
    rx := r shl 1;
    ry := 0;
    radio := radiox;

    while radio > 0 do
    begin

      if r > 0 then
      begin
        inc (y1);
        dec (y2);
        inc (ry, dx);
        dec (r, ry);
      end;

      if r <= 0 then
      begin
        dec (radio);
        inc (x1);
        dec (x2);
        dec (rx, dy);
        inc (r, rx);
      end;

      filtrapunto (x1, y1);
      filtrapunto (x1, y2);
      filtrapunto (x2, y1);
      filtrapunto (x2, y2);
    end;

  end;

  procedure filtraruedallena (x, y : integer; radiox, radioy : word);

    var
      radio,
      x1,
      x2,
      y1,
      y2 : integer;
      aq,
      bq,
      dx,
      dy,
      r,
      rx,
      ry : longint;

  begin
    filtralinea (x - radiox, y, x + radiox, y);
    x1 := x - radiox;
    y1 := y;
    x2 := x + radiox;
    y2 := y;
    aq := longint (radiox) * radiox;
    bq := longint (radioy) * radioy;
    dx := aq shl 1;
    dy := bq shl 1;
    r := radiox * bq;
    rx := r shl 1;
    ry := 0;
    radio := radiox;

    while radio > 0 do
    begin

      if r > 0 then
      begin
        inc (y1);
        dec (y2);
        inc (ry, dx);
        dec (r, ry);
      end;

      if r <= 0 then
      begin
        dec (radio);
        inc (x1);
        dec (x2);
        dec (rx, dy);
        inc (r, rx);
      end;

      filtralinea (x1, y1, x2, y1);
      filtralinea (x1, y2, x2, y2);
    end;

  end;

  procedure filtrasimbolos (simbolos : pchar; x, y : integer);

    var
      indicesimbolo : word;
      indicetabla : pointer;
      indicelinea,
      linea : byte;

  begin
    indicesimbolo := 0;

    while simbolos [indicesimbolo] <> #0 do
    begin
      indicetabla := ptr (seg (fundicion^), ofs (fundicion^) + word
        (simbolos [indicesimbolo]) * 8);

      for indicelinea := 0 to 7 do
      begin
        linea := byte (indicetabla^);
        if linea and 128 <> 0 then filtrapunto (x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 64 <> 0 then filtrapunto (x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 32 <> 0 then filtrapunto (x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 16 <> 0 then filtrapunto (x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 8 <> 0 then filtrapunto (x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 4 <> 0 then filtrapunto (x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 2 <> 0 then filtrapunto (x, y);
        inc (x, masx1simbolo);
        inc (y, masy1simbolo);
        if linea and 1 <> 0 then filtrapunto (x, y);
        inc (x, masx2simbolo);
        inc (y, masy2simbolo);
        inc (word (indicetabla));
      end;

      inc (x, masx3simbolo);
      inc (y, masy3simbolo);
      inc (indicesimbolo);
    end;

  end;

  procedure metefiltro;
  begin
    filtro := filtroborrosidad;
  end;

{/// Ratn /////////////////////////////////////////////////////////////////}

{Encabezado basado en la documentacin de Borland Pascal.}
  procedure atienderaton (flags, cs, ip, ax, bx, cx, dx, si, di, ds, es, bp :
    word); interrupt;
  begin
{Adems de flags, cs e ip; recibe:
  ax = llamada (la misma con que se instal)
  bx = estado de los botones
    binario 0 = 1 si botn izquierdo fue presionado, 0 si fue liberado
    binario 1 = 1 si botn derecho fue presionado, 0 si fue liberado
  cx = columna
  dx = fila
  di = movimiento horizontal
  si = movimiento vertical}

{Guarda los valores importantes.}
    xraton := cx;
    yraton := dx;
    botonizquierdoraton := bx and 1 <> 0;
    botonderechoraton := bx and 2 <> 0;
  end;

  procedure ajustaratonpagina;
  begin

    if umd. error = 0 then
    begin
      umd. ponerangox (x1pagina, x2pagina);
      umd. ponerangoy (y1pagina, y2pagina);
    end;

  end;

  procedure ajustaratonplantilla;
  begin

    if umd. error = 0 then
    begin
      umd. ponerangox (x1plantilla, x2plantilla);
      umd. ponerangoy (y1plantilla, y2plantilla);
    end;

  end;

  procedure centraraton;
  begin

    if umd. error = 0 then
    begin
      umd. mueve (xcentropagina, ycentropagina);
    end;

  end;

  procedure meteraton;
  begin

    if umd. error = 0 then
    begin
      ajustaratonpagina;
      centraraton;

{Instala la interrupcin que recibe valores del manejador cuando se mueve,
presiona o libera el ratn.}
      botonizquierdoraton := false;
      botonderechoraton := false;
      xraton := xcentropagina;
      yraton := ycentropagina;
      umd. cambiarutinas ($ffff, @atienderaton, viejallamadaraton,
        viejarutinaraton);
    end;

  end;

  procedure sacaraton;
  begin

    if umd. error = 0 then
    begin
{Quita la interrupcin que recibe valores del manejador.}
      umd. ponerutina (viejallamadaraton, viejarutinaraton);
    end;

  end;

{/// Principal /////////////////////////////////////////////////////////////}

{$f+}
  procedure sale;
{$f-}
  begin
    exitproc := viejoexitproc;

    sacaraton;

    ponemodo (viejomodo);
  end;

begin
  error := 0;

  if test8086 < 2 then
  begin
    error := 1;
  end;

  viejomodo := usvbe. tomamodo;
  usvbe. ponemodo (modo);

  if usvbe. tomamodo <> modo then
  begin
    error := 2;
  end;

  usvbe. informamodo (modo, informemodo);

{Asigna funciondirecta de usvbe para habilitar ponelugarventanadirecta y
tomalugarventanadirecta.}
  usvbe. funciondirecta := informemodo. funciondirecta;

  incrementoventana := (tamanoventana div unidadgranularidad) div
    informemodo. granularidad;
  lugarventana := tomalugarventana (ventana);

  metepagina;
  meteplantilla;
  metesimbolo;
  metefiltro;
  meteraton;

  viejoexitproc := exitproc;
  exitproc := @sale;
end.